# -*- coding: utf-8 -*-
"""
@author: Sam Schott  (ss2151@cam.ac.uk)

(c) Sam Schott; This work is licensed under a Creative Commons
Attribution-NonCommercial-NoDerivs 2.0 UK: England & Wales License.

The following APIs should remain stable for frontends:

* maestral.main.Maestral
* maestral.constants
* maestral.daemon
* maestral.errors
* maestral.utils.appdirs
* maestral.utils.autostart

"""

__version__ = '1.0.3'
__author__ = 'Sam Schott'
__url__ = 'https://github.com/SamSchott/maestral'
